import random
from typing import Union, Tuple


class RandomIDGenerator:
    def __init__(
        self, 
        start: int=0, 
        stop: int=65535
    ):
        self.data = [
            (start, stop)
        ]

    def generate(self):

        index = random.randrange(
            len(self.data)
        )

        rng = self.data[index]
        id = random.randrange(
            rng[0], 
            rng[1] + 1
        )

        rngs = []
        if id > rng[0]:
            rngs.append((rng[0], id - 1))

        if id < rng[1]:
            rngs.append((id + 1, rng[1]))

        self.data[index:index + 1] = rngs

        return id

    def put(self, value: int) -> None:

        size = len(self.data)

        for index, rng in enumerate(self.data):
            if value < rng[0]: 
                break
        
        else:
            index = size

        last_rng: Union[Tuple[int, int], None] = None
        next_rng: Union[Tuple[int, int], None] = None

        if index > 0:
            last_rng = self.data[index - 1]

        if index < size:
            next_rng = self.data[index]

        if last_rng is not None and last_rng[1] == value - 1:
            last_rng = last_rng[0], value

        if next_rng is not None and next_rng[0] == value + 1:
            next_rng = value, next_rng[1]

        has_last_range = last_rng is not None
        has_next_range = next_rng is not None

        if has_last_range and has_next_range and last_rng[1] == next_rng[0]:
            last_rng = last_rng[0], next_rng[1]
            next_rng = None

        rngs = []
        if last_rng is not None:
            rngs.append(last_rng)

        not_last_range = last_rng is None or last_rng[1] < value
        not_next_range = next_rng is None or value < next_rng[0]

        if not_last_range and not_next_range:
            rngs.append((value, value))

        if next_rng is not None:
            rngs.append(next_rng)

        start = max(0, index - 1)
        end = min(index + 1, size)
        
        self.data[start:end] = rngs



import asyncio
from typing import Callable, Tuple


class MercurySyncTCPServerProtocol(asyncio.Protocol):
    def __init__(
        self, 
        callback: Callable[
            [
                bytes,
                Tuple[str, int]
            ],
            bytes
        ]
    ):
        super().__init__()
        self.callback = callback
        self.transport: asyncio.Transport = None

    def connection_made(self, transport) -> str:
        self.transport = transport

    def data_received(self, data: bytes):
        self.callback(
            data,
            self.transport
        )


from __future__ import annotations
from enum import Enum
from typing import Dict, Optional
'''
Constants of DNS types.
'''

class RecordType(Enum):
    NONE = 0
    A = 1
    NS = 2
    CNAME = 5
    SOA = 6
    PTR = 12
    MX = 15
    TXT = 16
    AAAA = 28
    SRV = 33
    NAPTR = 35
    ANY = 255
        

class RecordTypesMap:

    def __init__(self) -> None:
        self.names_mapping: Dict[RecordType, str] = {}
        self.codes_mapping: Dict[RecordType, int] = {}
        self.types_by_code: Dict[int, RecordType] = {}
        self.types_by_name: Dict[str, RecordType] = {}

        for record_type in RecordType:
            self.names_mapping[record_type] = record_type.name
            self.codes_mapping[record_type] = record_type.value
            self.types_by_code[record_type.value] = record_type
            self.types_by_name[record_type.name] = record_type


    def get_name_by_code(
        self,
        code: int, 
        default: Optional[RecordType]=None
    ) -> str:
        record_type = self.types_by_code.get(
            code, 
            default
        )

        if record_type is None:
            return str(code)

        return record_type.name
    
    def get_code_by_name(
        self,
        name: str,
        default: Optional[RecordType]=None
    ):
        record_type = self.types_by_name.get(
            name, 
            default
        )

        if record_type is None:
            raise KeyError(f'No record type matches code - {name}')

        return record_type.value



from __future__ import annotations
import asyncio
import socket
import zstandard
from dtls import do_patch
from mercury_sync.connection.udp.protocols import MercurySyncUDPProtocol
from mercury_sync.env import Env
from typing import (
    Optional,
)
from .mercury_sync_udp_connection import MercurySyncUDPConnection

do_patch()


class MercurySyncUDPMulticastConnection(MercurySyncUDPConnection):
    """Implementation of Zeroconf Multicast DNS Service Discovery
    Supports registration, unregistration, queries and browsing.
    """
    def __init__(
        self, 
        host: str,
        port: int,
        instance_id: int,
        env: Env,
    ):
        super().__init__(
            host,
            port,
            instance_id,
            env
        )

        self._mcast_group = env.MERCURY_SYNC_MULTICAST_GROUP

        if self._mcast_group is None:
            self.group = ('', self.port)

        else:
            self.group = (self._mcast_group, self.port)

    async def connect_async(
        self, 
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None,
        worker_socket: Optional[socket.socket]=None
    ) -> None:
        
        self._loop = asyncio.get_event_loop()
        self._running = True

        self._semaphore = asyncio.Semaphore(self._max_concurrency)

        self._compressor = zstandard.ZstdCompressor()
        self._decompressor = zstandard.ZstdDecompressor()

        if worker_socket is None:
            self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
            self.udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            try:
                self.udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                self.udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            except Exception:
                pass

            self.udp_socket.setsockopt(socket.SOL_IP, socket.IP_MULTICAST_TTL, 255)
            self.udp_socket.setsockopt(socket.SOL_IP, socket.IP_MULTICAST_LOOP, 1)

            try:
                self.udp_socket.bind(self.group)
            except ConnectionRefusedError:
                pass

            except OSError:
                pass
            
            self.udp_socket.setsockopt(
                socket.SOL_IP, 
                socket.IP_MULTICAST_IF, 
                socket.inet_aton(self.host) + socket.inet_aton('0.0.0.0')
            )

            if self._mcast_group is not None:
                self.udp_socket.setsockopt(
                    socket.SOL_IP, 
                    socket.IP_ADD_MEMBERSHIP, 
                    socket.inet_aton(self.udp_socket) + socket.inet_aton('0.0.0.0')
                )

            self.udp_socket.setblocking(False)

        else:
            self.udp_socket = worker_socket


        if cert_path and key_path:
            self._udp_ssl_context = self._create_udp_ssl_context(
                cert_path=cert_path,
                key_path=key_path,
            )

            self.udp_socket = self._udp_ssl_context.wrap_socket(self.udp_socket)

        server = self._loop.create_datagram_endpoint(
            lambda: MercurySyncUDPProtocol(
                self.read
            ),
            sock=self.udp_socket
        )

        transport, _ = await server
        self._transport = transport

        self._cleanup_task = self._loop.create_task(self._cleanup())



from __future__ import annotations
import asyncio
import json
import socket
import ssl
import zstandard
import traceback
from collections import deque, defaultdict
from mercury_sync.env import Env
from mercury_sync.connection.base.connection_type import ConnectionType
from mercury_sync.models.http_message import HTTPMessage
from mercury_sync.models.http_request import HTTPRequest
from mercury_sync.models.request import Request
from mercury_sync.models.message import Message
from typing import Tuple, Union, Optional, Deque, Dict, List, Any
from .mercury_sync_tcp_connection import MercurySyncTCPConnection
from .protocols import MercurySyncTCPClientProtocol


class MercurySyncHTTPConnection(MercurySyncTCPConnection):

    def __init__(
        self, 
        host: str, 
        port: int,
        instance_id: int, 
        env: Env,
    ) -> None:
        super().__init__(
            host, 
            port, 
            instance_id, 
            env
        )

        self._waiters: Deque[asyncio.Future] = deque()
        self._connections: Dict[str, List[asyncio.Transport]] = defaultdict(list)
        self._http_socket: Union[socket.socket, None] = None
        self._hostnames: Dict[Tuple[str, int], str] = {}
        self._max_concurrency = env.MERCURY_SYNC_HTTP_POOL_SIZE

        self.connection_type = ConnectionType.HTTP
        self._is_server = env.MERCURY_SYNC_USE_HTTP_SERVER
        self._use_encryption = env.MERCURY_SYNC_USE_HTTP_MSYNC_ENCRYPTION

        self._supported_handlers: Dict[str, Dict[str, str]] = defaultdict(dict)

    async def connect_client(
        self,
        address: Tuple[str, int],
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None,
        worker_socket: Optional[socket.socket]=None,
        is_ssl: bool=False,
        hostname: str=None,
    ) -> None:
        
        self._hostnames[address] = hostname
        
        if self._semaphore is None:
            self._semaphore = asyncio.Semaphore(self._max_concurrency)

        if self._compressor is None and self._decompressor is None:
            self._compressor = zstandard.ZstdCompressor()
            self._decompressor = zstandard.ZstdDecompressor()
        
       
        if cert_path and key_path:
            self._client_ssl_context = self._create_client_ssl_context(
                cert_path=cert_path,
                key_path=key_path
            ) 

        elif is_ssl:
            self._client_ssl_context = self._create_general_client_ssl_context(
                cert_path=cert_path,
                key_path=key_path
            ) 

        last_error: Union[Exception, None] = None

        for _ in range(self._tcp_connect_retries):

            try:

                self._connections[address] = await asyncio.gather(*[
                    self._connect_client(
                        address,
                        hostname=hostname,
                        worker_socket=worker_socket
                    ) for _ in range(self._max_concurrency)
                ])

                return
            
            except ConnectionRefusedError as connection_error:
                last_error = connection_error

            await asyncio.sleep(1)

        if last_error:
            raise last_error

    def _create_general_client_ssl_context(
        self,
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None,
    ):
        ctx = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
        ctx.check_hostname = False
        ctx.verify_mode = ssl.CERT_NONE

        return ctx

        
    async def _connect_client(
        self,
        address: Tuple[str, int],
        hostname: str=None,
        worker_socket: Optional[socket.socket]=None,
    ) -> asyncio.Transport:
        
        self._loop = asyncio.get_event_loop()

        if worker_socket is None:

            http_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            http_socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            await self._loop.run_in_executor(None, http_socket.connect, address)

            http_socket.setblocking(False)

        else:
            http_socket = worker_socket
 
        transport, _ = await self._loop.create_connection(
            lambda: MercurySyncTCPClientProtocol(
                self.read
            ),
            sock=http_socket,
            server_hostname=hostname,
            ssl=self._client_ssl_context
        )

        return transport
    
    async def send(
        self, 
        event_name: str,
        data: HTTPRequest, 
        address: Tuple[str, int]
    ):
        async with self._semaphore:


            connections = self._connections.get(address)
            if connections is None:

                connections = await self.connect_client(
                    address,
                    cert_path=self._client_cert_path,
                    key_path=self._client_key_path,
                    is_ssl='https' in data.url
                )

                self._connections[address] = connections

            client_transport = connections.pop()

            result: Union[bytes, None] = None

            try:

                encoded_request = data.prepare_request()
                encrypted_request = self._encryptor.encrypt(encoded_request)
                compressed_request = self._compressor.compress(encrypted_request)

                client_transport.write(compressed_request)

                waiter = self._loop.create_future()
                self._waiters.append(waiter)
                
                result = await waiter

            except Exception:
                self._connections[address].append(
                    await self._connect_client(
                        (
                            self.host,
                            self.port
                        ),
                        hostname=self._hostnames.get(address)
                    )
                )

            self._connections[address].append(client_transport)

            return result
    
    async def send_request(
        self, 
        data: HTTPRequest, 
        address: Tuple[str, int]
    ):
        async with self._semaphore:

            encoded_request = data.prepare_request()

            connections = self._connections.get(address)
            client_transport = connections.pop()

            result: Union[bytes, None] = None

            try:

                client_transport.write(encoded_request)

                waiter = self._loop.create_future()
                self._waiters.append(waiter)
                
                result = await waiter

            except Exception:
                self._connections[address].append(
                    await self._connect_client(
                        (
                            self.host,
                            self.port
                        ),
                        hostname=self._hostnames.get(address)
                    )
                )

            self._connections[address].append(client_transport)

            return result

    def read(
        self, 
        data: bytes, 
        transport: asyncio.Transport
    ) -> None:
        
        if self._is_server:
            self._pending_responses.append(
                asyncio.create_task(
                    self._route_request(
                        data,
                        transport
                    )
                )
            )

        elif bool(self._waiters):

            waiter = self._waiters.pop()
            waiter.set_result(
                HTTPRequest.parse(data)
            )

    async def _route_request(
        self, 
        data: bytes,
        transport: asyncio.Transport
    ):
        
        if self._use_encryption:
            decompressed_data = self._decompressor.decompress(data)
            data = self._encryptor.decrypt(decompressed_data)
        
        request_data = data.split(b'\r\n')
        method, path, request_type = request_data[0].decode().split(' ')

        query: Union[str, None] = None
        if '?' in path:
            path, query = path.split('?')

        request = Request(
            path,
            method,
            query,
            request_data
        )

        try:
            
            handler = self.events[f'{method}_{path}']

            response_info: Tuple[
                Union[str, None],
                int
            ] = await handler(request)

            (
                response_data, 
                status_code
            ) = response_info

            head_line = f'HTTP/1.1 {status_code} OK'
            
            encoded_data: str = ''
            if response_data:
                data = response_data.encode()
                encoded_data = f'{data}\r\n'

                content_length = len(encoded_data)
                headers = f'content-length: {content_length}\r\n'

            else:
                headers = 'content-length: 0\r\n'

            if handler.response_headers:
                for key in handler.response_headers:
                    headers = f'{headers}{key}: {request.headers[key]}\r\n'

            transport.write(
                f'{head_line}\r\n{headers}\r\n{encoded_data}'.encode()
            )

            transport.close()

        except KeyError:
            
            if self._supported_handlers.get(request.path) is None:
            
                not_found_response = HTTPMessage(
                    path=request.path,
                    status=404,
                    error='Not Found',
                    protocol=request_type,
                    method=request.method
                )

                transport.write(not_found_response.prepare_response())
                transport.close()

            elif self._supported_handlers[request.path].get(request.method) is None:

                method_not_allowed_response = HTTPMessage(
                    path=request.path,
                    status=405,
                    error='Method Not Allowed',
                    protocol=request_type,
                    method=request.method
                )

                transport.write(method_not_allowed_response.prepare_response())
                transport.close()

        except Exception:
            server_error_respnse = HTTPMessage(
                path=request.path,
                status=500,
                error='Internal Error',
                protocol=request_type,
                method=request.method
            )

            transport.write(server_error_respnse.prepare_response())
            transport.close()





'''
Cache module.
'''

import json
import os
from urllib import request
from pathlib import Path

from mercury_sync.discovery.dns.core.record import (
    Record,
    RecordType,
    RecordTypesMap
)


__all__ = [
    'core_config',
    'get_name_cache',
    'get_root_servers',
]

CONFIG_DIR = os.environ.get('MERCURY_SYNC_DNS_CONFIG_DIR',
                            os.path.expanduser('~/.config/mercury_dns'))
os.makedirs(CONFIG_DIR, exist_ok=True)
CACHE_FILE = os.path.join(CONFIG_DIR, 'named.cache.txt')

try:
    with open(os.path.join(CONFIG_DIR, 'config.json')) as f:
        user_config = json.load(f)
except Exception:
    user_config = None

core_config = {
    'default_nameservers': [
        '8.8.8.8',
        '8.8.4.4',
    ],
}
if user_config is not None:
    core_config.update(user_config)
    del user_config


def get_nameservers():
    return []


def get_name_cache(
    url='ftp://rs.internic.net/domain/named.cache',
    filename=CACHE_FILE,
    timeout=10
):

    try:
        res = request.urlopen(url, timeout=timeout)

    except Exception:
        pass

    else:
        with open(filename, 'wb') as f:
            f.write(res.read())


def get_root_servers(filename=CACHE_FILE):

    if not os.path.isfile(filename):
        get_name_cache(filename=filename)

    if not os.path.isfile(filename):
        return
    for line in Path(filename).read_text().splitlines():

        if line.startswith(';'):
            continue

        parts = line.lower().split()
        if len(parts) < 4:
            continue

        name = parts[0].rstrip('.')

        types_map = RecordTypesMap()
        record_type = types_map.types_by_code.get(parts[2], RecordType.NONE)

        data_str = parts[3].rstrip('.')
        data = Record.create_rdata(record_type, data_str)
        yield Record(
            name=name,
            record_type=record_type,
            data=data,
            ttl=-1,
        )



class InvalidServiceURLError(Exception):

    def __init__(self, url: str) -> None:
        super().__init__(
            f'Err. - {url} does not match required patter (instance_name)._(service_name)._(udp|tcp).(domain_name)'
        )


from .a_record_data import ARecordData
from .aaaa_record_data import AAAARecordData
from .cname_record_data import CNAMERecordData
from .domain_record_data import DomainRecordData
from .mx_record_data import MXRecordData
from .naptr_record_data import NAPTRRecordData
from .ns_record_data import NSRecordData
from .ptr_record_data import PTRRecordData
from .record_data import RecordData
from .record_types import (
    RecordType,
    RecordTypesMap
)
from .soa_record_data import SOARecordData
from .srv_record_data import SRVRecordData
from .txt_record_data import TXTRecordData
from .unsupported_record_data import UnsupportedRecordData


import asyncio
import socket
from mercury_sync.connection.base.connection_type import ConnectionType
from mercury_sync.connection.tcp import (
    MercurySyncHTTPConnection,
    MercurySyncTCPConnection
)
from mercury_sync.connection.udp import MercurySyncUDPConnection
from mercury_sync.env import (
    Env, 
    RegistrarEnv, 
    load_env
)
from mercury_sync.env.time_parser import TimeParser
from mercury_sync.models.dns_message import DNSMessage
from mercury_sync.models.http_message import HTTPMessage
from mercury_sync.discovery.dns.core.url import URL
from typing import Optional, Tuple, Union, Dict


class DNSClient:

    def __init__(
        self,
        host: str,
        port: int,
        instance_id: str,
        env: Env
    ) -> None:
        
        registrar_env: RegistrarEnv = load_env(RegistrarEnv)
        
        self.host = host
        self.port = port
        self.instance_id = instance_id
        self.env = env

        self._client_config = (
            host,
            port + 2,
            instance_id,
            env
        )
        
        self._connection_types: Dict[
            ConnectionType,
            Union[
                MercurySyncUDPConnection,
                MercurySyncTCPConnection,
                MercurySyncHTTPConnection
            ]
        ] = {
            ConnectionType.UDP: lambda config: MercurySyncUDPConnection(*config),
            ConnectionType.TCP: lambda config: MercurySyncTCPConnection(*config),
            ConnectionType.HTTP: lambda config: MercurySyncHTTPConnection(*config),
        }

        self._client: Union[
            MercurySyncUDPConnection,
            MercurySyncTCPConnection,
            MercurySyncHTTPConnection,
            None
        ] = None

        self._client_types = {
            "udp": ConnectionType.UDP,
            "tcp": ConnectionType.TCP,
            "http": ConnectionType.HTTP
        }

        self.client_type = self._client_types.get(
            registrar_env.MERCURY_SYNC_RESOLVER_CONNECTION_TYPE
        )

        self._request_timeout = TimeParser(
            registrar_env.MERCURY_SYNC_RESOLVER_REQUEST_TIMEOUT
        ).time

        self._connections: Dict[Tuple[str, int], bool] = {}
        self.cert_paths: Dict[str, str] = {}
        self.key_paths: Dict[str, str] = {}


    async def connect_client(
        self,
        url: URL,
        cert_path: Optional[str]=None,
        key_path: Optional[str]=None,
        worker_socket: Optional[socket.socket]=None
    ):
        
        self.cert_paths[url.address] = cert_path
        self.key_paths[url.address] = key_path
        
        self._client: Union[
            MercurySyncUDPConnection,
            MercurySyncTCPConnection,
            MercurySyncHTTPConnection
        ] = self._connection_types.get(
            self.client_type
        )(self._client_config)

        if self._client.connection_type == ConnectionType.TCP:
            await self._client.connect_client(
                url.address,
                cert_path=cert_path,
                key_path=key_path,
                worker_socket=worker_socket
            )

        elif self._client.connection_type == ConnectionType.HTTP:
            await self._client.connect_client(
                url.address,
                is_ssl=url.is_ssl,
                hostname=url.host,
                worker_socket=worker_socket
            )

        else:
            await self._client.connect_async(
                cert_path=cert_path,
                key_path=key_path,
                worker_socket=worker_socket
            )

    async def send(
        self,
        event_name: str,
        data: DNSMessage,
        url: URL
    ):
        if url.is_msync:
                
            return await asyncio.wait_for(
                self._send_msync(
                    event_name,
                    data,
                    url
                ),
                timeout=self._request_timeout
            ) 

        else:

            return await asyncio.wait_for(
                self._send(
                    event_name,
                    data,
                    url
                ),
                timeout=self._request_timeout
            )
    async def _send(
        self,
        event_name: str,
        data: DNSMessage,
        url: URL
    ):
        
        if self._client is None:
            await self.connect_client(
                url
            )
        
        if self._client.connection_type == ConnectionType.TCP:
            response = await self._client.send_bytes(
                event_name,
                data.to_tcp_bytes(),
                url.address
            )

            return DNSMessage.parse(response)
        
        elif self._client.connection_type == ConnectionType.HTTP:
            response: HTTPMessage = await self._client.send_request(
                event_name,
                data.to_http_bytes(url.url),
                url.address
            )

            return DNSMessage.parse(response.data)
        
        else:

            response = await self._client.send_bytes(
                event_name,
                data.to_udp_bytes(),
                url.address
            )

            return DNSMessage.parse(response)
        
    async def _send_msync(
        self,
        event_name: str,
        data: DNSMessage,
        url: URL
    ):
        if self._client is None:
            await self.connect_client(
                url
            )
        
        if self._client.connection_type == ConnectionType.TCP:
            response = await self._client.send(
                event_name,
                data,
                url.address
            )

            return DNSMessage.parse(response)
        
        else:

            response = await self._client.send(
                event_name,
                data,
                url.address
            )

            return DNSMessage.parse(response)
            


from enum import Enum

class QueryType(Enum):
    REQUEST=0
    RESPONSE=1

    @classmethod
    def by_value(cls, value: int):
        value_map = {
            0: QueryType.REQUEST,
            1: QueryType.RESPONSE
        }

        return value_map.get(
            value,
            QueryType.REQUEST
        )



from __future__ import annotations
import struct
from typing import Dict, Iterable, Tuple
from .record_data import RecordData
from .record_types import RecordType
from .utils import (
    load_domain_name,
    pack_domain_name
)


class MXRecordData(RecordData):
    '''A record'''

    def __init__(self, *args):
        super().__init__(
            RecordType.MX,
            data=args
        )

        (
            preference,
            exchange
        ) = args

        self.preference = preference
        self.exchange = exchange

    def __repr__(self):
        return '<%s-%s: %s>' % (self.type_name, self.preference, self.exchange)

    @classmethod
    def load(
        cls, 
        data: bytes, 
        cursor_position: int, 
        size: int
    ) -> Tuple[int, MXRecordData]:
        
        preference, = struct.unpack(
            '!H', 
            data[cursor_position:cursor_position + 2]
        )

        cursor_position, exchange = load_domain_name(data, cursor_position + 2)

        return cursor_position, MXRecordData(preference, exchange)

    def dump(self, names: Dict[str, int], offset: int) -> Iterable[bytes]:

        preference = struct.pack(
            '!H', 
            self.preference
        )

        domain_name = pack_domain_name(
            self.exchange, 
            names, 
            offset + 4
        ) 

        record_data = [
            preference,
            domain_name
        ]

        for data in record_data:
            yield data




import functools
from typing import (
    Optional, 
    List, 
    Literal,
    Dict
)


def server(
    path: Optional[str]="/",
    methods: List[
        Literal[
            "GET",
            "HEAD",
            "OPTIONS",
            "POST",
            "PUT",
            "PATCH",
            "DELETE"
        ]
    ]=["GET"],
    response_headers: Optional[Dict[str, str]]=None,
    as_http=False
):

    def wraps(func):

        func.server_only = True
        func.path = path
        func.methods = methods
        func.as_http =as_http
        func.response_headers = response_headers

        @functools.wraps(func)
        def decorator(
            *args,
            **kwargs
        ):
            return func(*args, **kwargs)
        
        return decorator
    
    return wraps
